import os
import base64
from werkzeug.wrappers import Request, Response, request
from werkzeug.routing import Map, Rule
from werkzeug.exceptions import HTTPException, NotFound
from werkzeug.middleware.shared_data import SharedDataMiddleware
from werkzeug.debug import DebuggedApplication
from jinja2 import Environment, FileSystemLoader, TemplateNotFound
from ..components.sessions import session_manager
from ..components.blueprint import Blueprint
from werkzeug.urls import url_encode
from itsdangerous import URLSafeTimedSerializer
import json
import mimetypes
from urllib.parse import quote as url_quote, quote_plus as url_quote_plus, urlencode as url_encode
from werkzeug.utils import send_from_directory, safe_join
import importlib

class Zylo:
    def __init__(self, __name__=None):
        self.template_folder = 'views'
        self.url_map = Map()
        self.static_folder = "static"
        self.error_handlers = {}
        self.middlewares = []
        self.host = 'localhost'
        self.port = 8000
        self.debug = True
        self.secret_key = os.urandom(24)
        self.serializer = URLSafeTimedSerializer(base64.urlsafe_b64encode(self.secret_key))
        self.blueprints = []
        self.__name__ = __name__
        self.config = {}
        self.template_backend = 'zylo.backends.ZyloTemplates'
        self.template_folder = 'views'
        self.load_settings()

    def load_settings(self):
        try:
            settings_module = importlib.import_module('settings')
            templates_setting = getattr(settings_module, 'TEMPLATES', None)
            self.template_backend = templates_setting[0]['BACKEND'] if templates_setting else self.template_backend
            assert self.template_backend == 'zylo.backends.ZyloTemplates', "This backend isn't supported by Zylo."
            self.template_folder = templates_setting[0]['DIRS'][0] if templates_setting and templates_setting[0]['DIRS'] else self.template_folder
            self.host = getattr(settings_module, 'HOST', self.host)
            self.port = getattr(settings_module, 'PORT', self.port)
            self.debug = getattr(settings_module, 'DEBUG', self.debug)
        except ImportError:
            pass  
        except (IndexError, KeyError, AssertionError, ValueError) as e:
            raise ValueError("Invalid TEMPLATES setting in settings.py.") from e
        self.template_env = Environment(loader=FileSystemLoader(self.template_folder))

    def add_url_rule(self, rule, endpoint, handler, methods=['GET']):
        def view_func(request, **values):
            return handler(request, **values)
        self.url_map.add(Rule(rule, endpoint=endpoint, methods=methods))
        setattr(self, endpoint, view_func)

    def route(self, rule, methods=['GET']):
        def decorator(handler):
            self.add_url_rule(rule, handler.__name__, handler, methods)
            return handler
        return decorator

    def errorhandler(self, code):
        def decorator(handler):
            self.error_handlers[code] = handler
            return handler
        return decorator

    def use(self, middleware):
        self.middlewares.append(middleware)

    def config(self):
        return self.config

    def register_blueprint(self, blueprint):
        self.blueprints.append(blueprint)

    def handle_request(self, request):
        adapter = self.url_map.bind_to_environ(request.environ)
        try:
            endpoint, values = adapter.match()
            handler = getattr(self, endpoint)
            response = handler(request, **values)
        except NotFound as e:
            response = self.handle_error(404, e, request)
        except HTTPException as e:
            response = e
        return response

    def handle_error(self, code, error, request):
        handler = self.error_handlers.get(code)
        if handler:
            return handler(error, request)
        else:
            return error

    def wsgi_app(self, environ, start_response):
        request = Request(environ)
        for blueprint in self.blueprints:
            if request.path.startswith(blueprint.url_prefix):
                request.blueprint = blueprint
                response = blueprint.wsgi_app(environ, start_response)
                return response

        session_id = request.cookies.get('session_id')
        session_data = session_manager.load_session(session_id)
        request.session = session_data
        response = self.handle_request(request)
        session_id = session_manager.save_session(request.session)

        if isinstance(response, Response):
            response.set_cookie('session_id', session_id, secure=True, httponly=True)

        return response(environ, start_response)

    def __call__(self, environ, start_response):
        app = self.wsgi_app
        for middleware in reversed(self.middlewares):
            app = middleware(app)
        return app(environ, start_response)

    def run(self, host=None, port=None, debug=None, secret_key=None):
        if host is not None:
            self.host = host
        if port is not None:
            self.port = port
        if debug is not None:
            self.debug = debug
        if secret_key is not None:
            self.secret_key = secret_key

        if self.debug:
            app = DebuggedApplication(self, evalex=True)
        else:
            app = self

        from werkzeug.serving import run_simple
        run_simple(self.host, self.port, app, use_reloader=True)



from functools import wraps
from datetime import datetime, timedelta
from werkzeug.wrappers import Response


class Limiter:
    def __init__(self, app=None):
        self.limit_value = 100
        self.period = 60
        self.cache = {}
        self.app = app

    def limit(self, limit=100, period=60):
        def decorator(handler):
            @wraps(handler)
            def wrapped_handler(request, *args, **kwargs):
                key = request.remote_addr
                now = datetime.now()
                if key in self.cache:
                    requests = self.cache[key]
                    valid_requests = [
                        req for req in requests if (now - req) <= timedelta(seconds=period)
                    ]
                    if len(valid_requests) >= limit:
                        return Response('Rate limit exceeded', status=429)
                    valid_requests.append(now)
                    self.cache[key] = valid_requests
                else:
                    self.cache[key] = [now]
                return handler(request, *args, **kwargs)

            return wrapped_handler

        return decorator




import os
import base64
from cryptography.fernet import Fernet
from itsdangerous import URLSafeTimedSerializer


class SessionManager:
    def __init__(self, secret_key):
        self.secret_key = secret_key
        self.serializer = URLSafeTimedSerializer(secret_key)
        self.cipher_suite = Fernet(base64.urlsafe_b64encode(secret_key))

    def encode_data(self, data):
        serialized_data = self.serializer.dumps(data)
        encrypted_data = self.cipher_suite.encrypt(serialized_data.encode('utf-8'))
        return base64.urlsafe_b64encode(encrypted_data).decode('utf-8')

    def decode_data(self, data):
        encrypted_data = base64.urlsafe_b64decode(data.encode('utf-8'))
        decrypted_data = self.cipher_suite.decrypt(encrypted_data).decode('utf-8')
        return self.serializer.loads(decrypted_data)

    def save_session(self, session_data):
        session_id = self.encode_data(session_data)
        return session_id

    def load_session(self, session_id):
        try:
            session_data = self.decode_data(session_id)
        except Exception:
            session_data = {}
        return session_data


# Create a global instance of the session manager
secret_key = os.urandom(32)
session_manager = SessionManager(secret_key)



from werkzeug.routing import Map, Rule
from werkzeug.exceptions import NotFound
from werkzeug.wrappers import Request, Response
from .sessions import session_manager

class Blueprint:
    def __init__(self, name, import_name, url_prefix=''):
        self.name = name
        self.import_name = import_name
        self.url_prefix = url_prefix
        self.url_map = Map()
        self.error_handlers = {}
        self.session_manager = session_manager

    def add_url_rule(self, rule, endpoint, handler, methods=['GET']):
        rule = self.url_prefix + rule
        self.url_map.add(Rule(rule, endpoint=endpoint, methods=methods))
        setattr(self, endpoint, handler)

    def route(self, rule, methods=['GET']):
        def decorator(handler):
            self.add_url_rule(rule, handler.__name__, handler, methods)
            return handler

        return decorator

    def errorhandler(self, code):
        def decorator(handler):
            self.error_handlers[code] = handler
            return handler

        return decorator

    def handle_request(self, request):
        adapter = self.url_map.bind_to_environ(request.environ)
        try:
            endpoint, values = adapter.match()
            handler = getattr(self, endpoint)
            response = handler(request, **values)
        except NotFound as e:
            response = self.handle_error(404, e)
        return response

    def handle_error(self, code, error):
        handler = self.error_handlers.get(code)
        if handler:
            return handler(error)
        else:
            response = Response(str(error), status=code)
            response.set_cookie('session_id', '', expires=0)
            return response


    def wsgi_app(self, environ, start_response):
        request = Request(environ)
        session_id = request.cookies.get('session_id')
        session_data = self.session_manager.load_session(session_id)
        request.session = session_data
        response = self.handle_request(request)
        session_id = self.session_manager.save_session(request.session)
        response.set_cookie('session_id', session_id, secure=True, httponly=True)
        return response(environ, start_response)

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)
